#include <math.h>
#include <mem.h>
#include <stdio.h>
#include "wrdef.h"
#include "wrio.h"

typedef DWORD MATRIX[2][2];

static int GetVcoParams(int hRadio, DWORD fvco, DWORD fref, int minR, int maxR, UINT *R, DWORD *Ntot)
{
	double Kn, Kp, af;
	MATRIX M, tM, cM;
	int i;
	UINT tR;
	DWORD tN, raf;

	Kn = (double)fvco / fref;
	M[0][0] = 0;
	M[0][1] = 1;
	M[1][0] = 1;
	M[1][1] = (long)Kn;
	Kp = modf(Kn, &Kn);
	memcpy(tM, M, sizeof(MATRIX));
    while ((Kp > 1e-6) && (M[0][1] < maxR)) 
	{
        Kn = 1 / Kp;
        tM[1][1] = (long)Kn;
        memcpy(cM, M, sizeof(MATRIX));
        M[0][0] = cM[0][0] * tM[0][0] + cM[0][1] * tM[1][0]; // M = cM * tM - matrix multiplication
        M[0][1] = cM[0][0] * tM[0][1] + cM[0][1] * tM[1][1];
        M[1][0] = cM[1][0] * tM[0][0] + cM[1][1] * tM[1][0];
        M[1][1] = cM[1][0] * tM[0][1] + cM[1][1] * tM[1][1];
	    Kp = modf(Kn, &Kn);
    }
    if (M[0][1] > maxR)
		memcpy(M, cM, sizeof(MATRIX));
    *R = M[0][1];
    while (*R < minR) 
		*R += M[0][1];
    *Ntot = M[1][1] * *R / M[0][1];
    Kp = (double)*Ntot / *R;
    af = Kp * fref;
    raf = (long)(af + 0.5) - fvco;
	if (labs(raf) > 50)
		for (i = 1; i < 25; i++)
		{
            if (modf(Kp * i + 1e-10, &Kn) < 1e-6) 
			{
                Kn = 50 / i;
                if (raf > 0)
                    GetVcoParams(hRadio, (long)(af - Kn), fref, minR, maxR, &tR, &tN);
                else
                    GetVcoParams(hRadio, (long)(af + Kn + 1.0), fref, minR, maxR, &tR, &tN);
                fvco = (long)(tN / tR * fref + 0.5) - fvco;
                if (abs(raf) > labs(fvco)) 
				{
                    *R = tR;
                    *Ntot = tN;
                    raf = fvco;
                }
                break;
            }
        }
    return raf;
}



BOOL SetFreq1000(int hRadio, double freq, double *actfreq)
{
	static const double VcoOffsets[5] = {556.325e6, 249.125e6, 58.075e6, -249.125e6, -556.325e6};
	static const double FreqRanges[9] = {1.8e6, 30e6, 50e6, 118e6, 300e6, 513e6, 798e6, 1106e6, 21e9};
	static const double SmFilters[8] = {0.5e6, 0.95e6, 1.8e6, 3.5e6, 7.0e6, 14e6, 30e6, 118e6};

	static const BYTE l0data[2][9] = {
		{0x54, 0x54, 0x54, 0x58, 0x58, 0x98, 0xC0, 0x80, 0x40},
		{0x54, 0x54, 0x54, 0x54, 0x58, 0x98, 0xD0, 0x80, 0x40}};

	LPRADIOSETTINGS rs;
	DWORD Ntot;
	UINT R;
	double fvco, vcoofs;
	BYTE l0;
	int l0idx, range;
	BOOL PllOverride;
	int bufidx = 0;
	BYTE buf[16];

	if (!ValidateHandle(hRadio, &rs))
		return FALSE;

	PllOverride = (!(rs->riInfo.dwFeatures & RIF_USVERSION) &&
		((rs->riInfo.wHWVer == RHV_1500) || (rs->riInfo.wHWVer >= RHV_3100)));

	if (actfreq)
		*actfreq = freq;

	l0idx = (rs->riInfo.wHWVer < RHV_3000) ? 0 : 1;

	l0 = 0xf3;				// get current state of latch 0
	if (!McuTransfer(hRadio, 1, &l0, 1, &l0))
		return FALSE;

	for (range = 0; range < 9; range++)
		if (freq < FreqRanges[range])
			break;

	if (range > 4)
		vcoofs = VcoOffsets[range-4];
	else
		vcoofs = VcoOffsets[0];

	fvco = freq + vcoofs;

	// calculate PLL register values

	if (rs->iCurMode == RMD_FMW)
	{
		R = rs->dwRefFreq / 20000;
		Ntot = (long)(fvco / 20e3 + 0.5);	// get closest freq to multiple of 20 kHz in FMW
	} else
		GetVcoParams(hRadio, (long)(fvco + 0.5), rs->dwRefFreq, rs->dwRefFreq / 20000, rs->dwRefFreq / 5000, &R, &Ntot);

	if (PllOverride)
		buf[bufidx++] = 0x58;	// activate -SETPLL line on non-US 1500s & 3100s

	// set PLL C register
	buf[bufidx++] = 0x6d;
	buf[bufidx++] = 0x2c;

	// calculate actual VCO frequency
	fvco = (double)Ntot / R * rs->dwRefFreq - vcoofs;

	if (PllOverride)	// toggle -SETPLL line
	{
		buf[bufidx++] = 0x59;
		buf[bufidx++] = 0x58;
	}

	// set PLL R register
	R |= 0x4000;
	buf[bufidx++] = 0x6e;
	buf[bufidx++] = R & 0xff;
	buf[bufidx++] = R >> 8;

	if (PllOverride)	// toggle -SETPLL line
	{
		buf[bufidx++] = 0x59;
		buf[bufidx++] = 0x58;
	}

	// set PLL A register
	Ntot = 0x300000L + (Ntot & 63) + ((Ntot >> 6) << 8);
	buf[bufidx++] = 0x6f;
	buf[bufidx++] = Ntot & 0xff;
	buf[bufidx++] = (Ntot >> 8) & 0xff;
	buf[bufidx++] = Ntot >> 16;

	if (PllOverride)	// clear -SETPLL line
		buf[bufidx++] = 0x59;

	if (rs->riInfo.dwFeatures & RIF_USVERSION)
		buf[bufidx++] = 0x68;	// initiate PLL programming if US version

	// set mixer and band lines (latch 0)
	buf[bufidx++] = 0xf0;
	buf[bufidx++] = (l0 & 0x23) | l0data[l0idx][range];

	if (!McuTransfer(hRadio, bufidx, buf, 0, NULL))
		return FALSE;

	if (rs->riInfo.dwFeatures & RIF_USVERSION)
	{
		// if US version, check to see if PLL/latch 0 combination was valid
		l0 = 0x93;
		if (!McuTransfer(hRadio, 1, &l0, 1, &l0) || l0)
			return FALSE;		// invalid PLL params if l0 is non-zero
	}

	// if Spectrum Monitor series, program HF filter board
	if ((rs->riInfo.wHWVer >= RHV_3000) &&
		(rs->riInfo.wHWVer <= RHV_3100))
	{
		if (freq < 118e6)
		{
			for (range = 0; range < 8; range++)
				if (freq < SmFilters[range])
					break;

			if (range & 1)
				buf[0] = 0x54;
			else
				buf[0] = 0x55;

			if (range & 2)
				buf[1] = 0x52;
			else
				buf[1] = 0x53;

			if (range & 4)
				buf[2] = 0x77;
			else
				buf[2] = 0x76;
		}
		else
		{
			buf[0] = 0x77;

			if (freq >= 1106e6)
			{
				buf[1] = 0x53;
				buf[2] = 0x54;
			}
			else
				if (freq >= 798e6)
				{
					buf[1] = 0x52;
					buf[2] = 0x55;
				}
				else
				{
					buf[1] = 0x52;
					buf[2] = 0x54;
				}
		}
		if (!McuTransfer(hRadio, 3, buf, 0, NULL))
			return FALSE;
	}

	rs->ftActFreq = fvco;
	if (actfreq)
		*actfreq = fvco;

	return TRUE;
}


BOOL SetFreq2000(int hRadio, double freq, double *actfreq)
{
	static const BYTE ShfData[3][4] = { {0x04, 0xa3, 0x99, 0x01},
										{0x02, 0xa3, 0x99, 0x86},
										{0x02, 0x71, 0x99, 0x86} };
	static const double FreqRanges[7] = {0.95e6, 1.8e6, 30e6, 118e6, 400e6, 1114e6, 21e9};
	static const BYTE l0data[7] = {0x4c, 0x4c, 0x4c, 0x4c, 0x4c, 0x80, 0x40};
	static const BYTE l1data[7] = {0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0xa0};

	LPRADIOSETTINGS rs;
	DWORD Ntot;
	UINT R;
	double fvco, vcoofs, tunedfreq;
	BYTE l0 = 0xf3, l1 = 0xf4;
	int range, newShfState;
	BYTE buf[16];

	if (!ValidateHandle(hRadio, &rs))
		return FALSE;

	if (actfreq)
		*actfreq = freq;

	tunedfreq = rs->ftFreqHz;
	newShfState = rs->iShfState;

	if (rs->riInfo.dwMaxFreqkHz > 2000000L)
	{
		if (freq > 2.6e9)
		{
			freq -= 2.5e9;
			tunedfreq -= 2.5e9;
			newShfState = 2;
		}
		else
		if (freq > 1.6e9) {
			freq = 2.7e9 - freq;
			tunedfreq = 2.7e9 - tunedfreq;
			newShfState = 1;
		}
		else
			newShfState = 0;
	}

	printf("SetFreq2000(%.0lf): ", freq);


	if (!McuTransfer(hRadio, 1, &l0, 1, &l0))	// get current state of latch 0
		return FALSE;
	if (!McuTransfer(hRadio, 1, &l1, 1, &l1))   // get current state of latch 1
		return FALSE;

	for (range = 0; range < sizeof(l0data); range++)
		if (tunedfreq < FreqRanges[range])
			break;

	if (tunedfreq < 400e6)
		vcoofs = 556.325e6;
	else if (tunedfreq < 808e6)
		vcoofs = 249.125e6;
	else if (tunedfreq < 1114e6)
		vcoofs = -249.125e6;
	else
		vcoofs = -556.325e6;

	fvco = freq + vcoofs;

	// calculate PLL register values

	if (rs->iCurMode == RMD_FMW)
	{
		R = rs->dwRefFreq / 20000;
		Ntot = (long)(fvco / 20e3 + 0.5);	// get closest freq to multiple of 20 kHz in FMW
	} else
		GetVcoParams(hRadio, (long)(fvco + 0.5), rs->dwRefFreq, rs->dwRefFreq / 20000, rs->dwRefFreq / 5000, &R, &Ntot);

	// calculate actual VCO frequency
	fvco = (double)Ntot / R * rs->dwRefFreq - vcoofs;

	printf("Ntot = %ld, R = %d, ", Ntot, R);

	// set PLL C register
	buf[0] = 0x6d;
	buf[1] = 0x2c;

	// set PLL R register
	R |= 0x4000;
	buf[2] = 0x6e;
	buf[3] = R & 0xff;
	buf[4] = R >> 8;

	// set PLL A register
	Ntot = 0x300000L + (Ntot & 63) + ((Ntot >> 6) << 8);
	buf[5] = 0x6f;
	buf[6] = Ntot & 0xff;
	buf[7] = (Ntot >> 8) & 0xff;
	buf[8] = Ntot >> 16;

	// set mixer and band lines (latch 0 & 1)
	buf[9] = 0xf0;
	buf[10] = (l0 & 0x23) | l0data[range];
	buf[11] = 0xf1;
	buf[12] = (l1 & 0x1f) | l1data[range];

	if (!McuTransfer(hRadio, 13, buf, 0, NULL))
		return FALSE;

	if (rs->riInfo.dwFeatures & RIF_USVERSION)
	{
		// if US version, check to see if PLL/latch 0 combination was valid
		l0 = 0x93;
		if (!McuTransfer(hRadio, 1, &l0, 1, &l0) || l0)
			return FALSE;	// invalid PLL params if l0 is non-zero
	}

	// update microwave convertor if applicable
	if ((rs->iShfState < 0) && (newShfState > 0))
		SendI2CData(hRadio, (rs->riInfo.dwMaxFreqkHz > 2600000) ? 0xc0 : 0xc2, 4, (LPBYTE)ShfData[0]);
	if (rs->iShfState != newShfState)
	{
		rs->iShfState = newShfState;
		SendI2CData(hRadio, (rs->riInfo.dwMaxFreqkHz > 2600000) ? 0xc0 : 0xc2, 4, (LPBYTE)ShfData[newShfState]);
	}
	if (newShfState == 1)
		fvco = 2.7e9 - fvco;
	else if (newShfState == 2)
		fvco += 2.5e9;

	rs->ftActFreq = fvco;
	if (actfreq)
		*actfreq = fvco;

	return TRUE;
}

