#include <dos.h>
#include <stdio.h>
#include <mem.h>

#include "wrdef.h"
#include "wrio.h"
#include "wrserial.h"

//----------------------------------------------------------------------
//
// This file contains platform specific code (PC in this case) for
// accessing WiNRADiO ISA devices. The serial part is generic (see
// wrserial.c to modify access to serial WiNRADiOs in different platforms).
//
// The only functions that may require modification to either remove or
// modify the access to the ISA devices are:
//     WriteMcuByte, ReadMcuByte, GetMcuStatus and PerformReset.
//

//----------------------------------------------------------------------
// Main function for sending a byte to the receiver's MCU
//----------------------------------------------------------------------


BOOL WriteMcuByte(int hRadio, BYTE c)
{
	DWORD timeout;

	if (RadioSettings[hRadio]->riInfo.iHWInterface == RHI_SERIAL)
	{
		// send byte to serial device (does not require modification)

		BYTE ack;

		while (ReadSerialByte(hRadio, &ack));	// flush read queue

		printf("%.2x, ", c);

		if (!SendSerialByte(hRadio, c))
			return FALSE;

		// have to wait until the acknowledge byte arrives
		timeout = GetTickCount + 200;
		while (!ReadSerialByte(hRadio, &ack))
			if (GetTickCount > timeout)
				return FALSE;

		ack = ~ack & 0xff;
		return (ack == c);	// received byte should be inverse of byte sent
	}
	else
	{
		// send byte to ISA device (may require modification or removal)

		WORD ioAddr = RadioSettings[hRadio]->wIoAddr;

		if (inp(ioAddr + 1) & 0x01) // empty read port if full
			inp(ioAddr);

		// MCU should be available within 10ms - timeout in 100ms
		timeout = GetTickCount + 100;
		while (inp(ioAddr + 1) & 0x02)
		{
			if (GetTickCount > timeout)
				return FALSE;
		}

		// send byte to MCU when the write port is empty
		outp(ioAddr, c);
	}
	return TRUE;
}

//----------------------------------------------------------------------
// Main function for reading a byte from the receiver's MCU
//----------------------------------------------------------------------

BOOL ReadMcuByte(int hRadio, PBYTE c)
{
	DWORD timeout;

	if (RadioSettings[hRadio]->riInfo.iHWInterface == RHI_SERIAL)
	{
		// read byte from serial device (does not require modification)

		// have to wait until a byte is available in the queue
		timeout = GetTickCount + 200;
		while (!ReadSerialByte(hRadio, c))
			if (GetTickCount > timeout)
				return FALSE;
	}
	else
	{
		// read byte from ISA device (may require modification or removal)

		WORD ioAddr = RadioSettings[hRadio]->wIoAddr;

		// byte should be available within 10ms - timeout in 100ms
		timeout = GetTickCount + 100;
		while (!(inp(ioAddr + 1) & 0x01))
		{
			if (GetTickCount > timeout)
				return FALSE;
		}

		// read byte from MCU when read port is full
		*c = inp(ioAddr);
	}
	return TRUE;
}

BOOL McuTransfer(int hRadio, int tc, PBYTE tb, int rc, PBYTE rb)
{
	// send all bytes in buffer to MCU
	while (tc--)
		if (!WriteMcuByte(hRadio, *tb++))
			return FALSE;

	// fill buffer with bytes read from the MCU
	while (rc--)
		if (!ReadMcuByte(hRadio, rb++))
			return FALSE;

	return TRUE;
}

BYTE GetMcuStatus(int hRadio)
{
	if (RadioSettings[hRadio]->riInfo.iHWInterface == RHI_SERIAL)
		return 0x20;	// PLL is locked and DSP doesn't exist
	else
		return inp(RadioSettings[hRadio]->wIoAddr + 1);
}

BOOL ValidateHandle(int hRadio, LPRADIOSETTINGS *lpRadioSettings)
{
	if ((hRadio > 0) && (hRadio <= MAX_DEVICES))
	{
		if (lpRadioSettings)
			*lpRadioSettings = RadioSettings[hRadio];
		return (RadioSettings[hRadio] != NULL);
	}
	return FALSE;
}

BOOL PerformReset(int hRadio)
{
	BYTE a, b;

	if (RadioSettings[hRadio]->riInfo.iHWInterface == RHI_SERIAL)
	{
		// initiates comms with serial devices (no mods required)

		WriteMcuByte(hRadio, 0);	// sychronise comms
		WriteMcuByte(hRadio, 0);
		WriteMcuByte(hRadio, 0);
		WriteMcuByte(hRadio, 0);

		// establish existance of WiNRADiO
		if (!(WriteMcuByte(hRadio, 0x0d) && ReadMcuByte(hRadio, &a) &&
			ReadMcuByte(hRadio, &b) && (a == 0x55) && (b == 0xaa)))
		{
			Delay(200);
			WriteMcuByte(hRadio, 0);
			WriteMcuByte(hRadio, 0);
			WriteMcuByte(hRadio, 0);
			WriteMcuByte(hRadio, 0);

			if (!(WriteMcuByte(hRadio, 0x0d) && ReadMcuByte(hRadio, &a) &&
				ReadMcuByte(hRadio, &b) && (a == 0x55) && (b == 0xaa)))
				return FALSE;
		}

		// increase baud rate to 38.4k
		WriteMcuByte(hRadio, 0xae);
		WriteMcuByte(hRadio, 3);	// = 152000 / baud rate

		SetBaudRate(hRadio, 38400l);

		return TRUE;
	}
	else
	{
		// reset the ISA WiNRADiO device (may require modification or removal)

		inp(RadioSettings[hRadio]->wIoAddr);	// clear outut buffer
		inp(RadioSettings[hRadio]->wIoAddr);

		outp(RadioSettings[hRadio]->wIoAddr + 1, 1);	// activate reset line

		Delay(100);

		outp(RadioSettings[hRadio]->wIoAddr + 1, 0);	// unreset MCU

		Delay(100);

		if (ReadMcuByte(hRadio, &a) && (a == 0x55))
			return TRUE;

		return (ReadMcuByte(hRadio, &a) && (a == 0x55));
	}
}

void Delay(int ms)	// generic delay function (platform independant)
{
	DWORD timeout = GetTickCount + ms;

	do ; while (GetTickCount <= timeout);
}

BOOL LockDetect(int hRadio)
{
	LPRADIOINFO ri = &RadioSettings[hRadio]->riInfo;
	BYTE b;

	if (ri->iHWInterface != RHI_ISA)
	{
		if ((ri->wHWVer <= RHV_1500) || ((ri->wHWVer >= RHV_3000) && (ri->wHWVer <= RHV_3100)))
			return TRUE;	// these externals do not support lock detect

		WriteMcuByte(hRadio, 0x43);
		ReadMcuByte(hRadio, &b);
		return (b < 127);
	}
	else
		return (GetMcuStatus(hRadio) & 4) != 0;
}

#define I2C_ENABLE	0x04
#define I2C_SCL		0x10
#define I2C_SDA 	0x80

static BOOL WriteToL2(int hRadio, BYTE b)
{
	return WriteMcuByte(hRadio, 0xf2) && WriteMcuByte(hRadio, b);
}

static BOOL SendI2CByte(int hRadio, BYTE b, BYTE raw)
{
	BOOL result;
	int i = 0x80;	// start with MSB
	BYTE t;
	DWORD timeout;

	// clear SCL and SDA lines
	WriteToL2(hRadio, raw);

	do
	{
		t = (b & i) ? raw | I2C_SDA : raw;
		WriteToL2(hRadio, t);
		WriteToL2(hRadio, t | I2C_SCL);
		WriteToL2(hRadio, t);
		i >>= 1;
	} while (i);

	// send ACK bit to wait for acknowledgement
	WriteToL2(hRadio, raw | I2C_SDA);
	WriteToL2(hRadio, raw | I2C_SDA | I2C_SCL);

	// if SDA is not low, then error sending data
	result = LockDetect(hRadio);

	WriteToL2(hRadio, raw | I2C_SDA);	// clear SCL

	// SDA becomes high when the IC receiver has acknowledged
	timeout = GetTickCount + 200;	// wait up to 200 ms
	do
		if (GetTickCount > timeout)
			return FALSE;
	while (!LockDetect(hRadio));

	return result;
}

BOOL SendI2CData(int hRadio, int iAddr, int cbSize, PBYTE lpData)
{
	BYTE raw;
	BOOL result = FALSE;

	WriteMcuByte(hRadio, 0xf5);
	ReadMcuByte(hRadio, &raw);
	raw &= 0x4b;

	WriteToL2(hRadio, raw | I2C_SCL | I2C_SDA);	// set SDA and SCL

	raw |= I2C_ENABLE;	// enable IC bus for duration of procedure
	WriteToL2(hRadio, raw | I2C_SCL | I2C_SDA);

	WriteToL2(hRadio, raw | I2C_SCL);	// initiate START condition

	// send IC address
	result = SendI2CByte(hRadio, iAddr, raw);

	// send IC data
	while ((cbSize-- > 0) && (result))
		result = SendI2CByte(hRadio, *lpData++, raw);

	WriteToL2(hRadio, raw | I2C_SCL);	// clear SDA and set SCL
	WriteToL2(hRadio, raw | I2C_SCL | I2C_SDA);	// send STOP condition

	WriteToL2(hRadio, raw & 0x8b | I2C_SCL | I2C_SDA);

	return result;
}
